package tv.smartclip.smartclientandroid.app.reference

import android.app.AlertDialog
import android.net.Uri

import android.os.Bundle
import android.view.View
import android.view.ViewGroup
import com.google.android.exoplayer2.*
import com.google.android.exoplayer2.source.ProgressiveMediaSource
import com.google.android.exoplayer2.source.TrackGroupArray
import com.google.android.exoplayer2.trackselection.TrackSelectionArray
import com.google.android.exoplayer2.ui.PlayerView
import com.google.android.exoplayer2.upstream.DataSource
import com.google.android.exoplayer2.upstream.DefaultBandwidthMeter
import com.google.android.exoplayer2.upstream.DefaultDataSourceFactory
import com.google.android.exoplayer2.util.Util
import dev.zieger.utils.log.Log
import kotlinx.android.synthetic.main.activity_in_stream.*
import kotlinx.coroutines.*
import tv.smartclip.smartclientandroid.lib.SxSequencer
import tv.smartclip.smartclientandroid.lib.dto.SxConfiguration
import tv.smartclip.smartclientandroid.lib.dto.SxLayoutConfiguration
import tv.smartclip.smartclientandroid.lib.dto.sequencer.SxSequencerAdSlot
import tv.smartclip.smartclientandroid.lib.dto.sequencer.SxSequencerRelativePosition
import tv.smartclip.smartclientandroid.lib.utils.asUnit
import tv.smartclip.smartclientandroid.lib.utils.toGone
import tv.smartclip.smartclientandroid.lib.video_player.*
import kotlin.coroutines.CoroutineContext
import kotlin.math.max
import kotlin.properties.Delegates

/**
 * Displays a video in a SimpleExoPlayer that plays after an advertisement's video.
 * Uses a custom [SxInstreamVideoPlayerWrapper].
 *
 * This class shows how to use the [SxSequencer] with any video player. We still using the
 * ExoPlayer here, but it can be replaced with your desired video player.
 */
class AbsVideoPlayerWrapperActivity : BaseActivity() {

    companion object {

        /**
         * Custom implementation of [SxInstreamVideoPlayerWrapper].
         *
         * Is needed when another video player than ExoPlayer should be used or the
         * [InstreamExoWrapper] does not fulfill your needs.
         *
         * In general this class connects the in and output events of the video player to the
         * corresponding properties defined in [SxVideoPlayerWrapper]. Because
         * [AbsVideoPlayerWrapper] already connects these properties to a delegate, setting their
         * value is enough to notify the underlying framework.
         *
         * @property playerView [View] of the video player.
         * @property contentVideoUrl Url of the content video.
         * @param userAgent UserAgent that should be used for the media file request.
         */
        class ReferenceVideoPlayerWrapper(private val playerView: PlayerView,
                                          private val contentVideoUrl: String,
                                          userAgent: String) : AbsVideoPlayerWrapper(),
                SxInstreamVideoPlayerWrapper {

            private var isShowingContent: Boolean = false
            private var lastContentVideoUrl: String? = null
            private var lastContentVideoWindowIndex: Int? = null
            private var lastContentVideoPosition: Long? = null

            override val overlayContainer: ViewGroup = playerView.adViewGroup
            override val surfaceView: View = playerView.videoSurfaceView
                    ?: throw IllegalArgumentException("VideoSurfaceView can not be null")

            private val context = overlayContainer.context
            private val dataSourceFactory: DataSource.Factory =
                    DefaultDataSourceFactory(context, Util.getUserAgent(context, userAgent),
                            DefaultBandwidthMeter.Builder(context).build())
            private val player = SimpleExoPlayer.Builder(context).build()

            /**
             * Executes the provided [block] on the Android main thread. Will send an
             * [PlaybackState.ERROR] event for any thrown [Exception] inside this [block].
             */
            private inline fun mainLaunch(crossinline block: suspend CoroutineScope.() -> Unit): Job =
                    mainScope.launch {
                        try {
                            block()
                        } catch (throwable: Throwable) {
                            event = SxVideoPlayerEvent(PlaybackState.ERROR(throwable), position, duration, mediaUrl)
                        }
                    }

            private val mainScope: CoroutineScope = object : CoroutineScope {
                override val coroutineContext: CoroutineContext = Dispatchers.Main + Job()
            }

            private val exoListener = object : Player.EventListener {

                override fun onPlayerStateChanged(playWhenReady: Boolean, playbackState: Int) {
                    updateJobActive = playWhenReady
                    update()
                }

                override fun onPlayerError(error: ExoPlaybackException) = mainLaunch {
                    Log.e("$error")
                    event = SxVideoPlayerEvent(PlaybackState.ERROR(error), player.currentPosition,
                            player.duration, mediaUrl)
                }.asUnit()

                // Not implementing empty listeners will result in crash of outstream use case with release version
                override fun onPlaybackParametersChanged(playbackParameters: PlaybackParameters) = Unit
                override fun onSeekProcessed() = Unit
                override fun onTracksChanged(trackGroups: TrackGroupArray, trackSelections: TrackSelectionArray) = Unit
                override fun onLoadingChanged(isLoading: Boolean) = Unit
                override fun onPositionDiscontinuity(reason: Int) = Unit
                override fun onRepeatModeChanged(repeatMode: Int) = Unit
                override fun onShuffleModeEnabledChanged(shuffleModeEnabled: Boolean) = Unit
                override fun onPlaybackSuppressionReasonChanged(playbackSuppressionReason: Int) = Unit
                override fun onTimelineChanged(timeline: Timeline, reason: Int) = Unit
                override fun onTimelineChanged(timeline: Timeline, manifest: Any?, reason: Int) = Unit
                override fun onIsPlayingChanged(isPlaying: Boolean) = Unit
            }

            /**
             * The update job will poll the video playback position when playback is active. This
             * needs to be done with a fast interval (at least 30 times per second). If your video
             * player offers a callback, that notifies you about playback position changes, you
             * should use this instead.
             */
            private var updateJob: Job? = null
            private var updateJobActive: Boolean by Delegates.observable(false) { _, _, new ->
                updateJob?.cancel()
                if (new)
                    updateJob = mainLaunch {
                        while (isActive) {
                            update()
                            delay(20)
                        }
                    }
            }

            init {
                playerView.useController = false
                playerView.player = player
                player.addListener(exoListener)

                // Using the control callback of the [playWhenReady]'s controllable will notify us
                // when [playWhenReady] changes.
                playWhenReadyControllable.control { play ->
                    mainLaunch {
                        player.playWhenReady = play
                        updateJobActive = play
                    }
                }
            }

            override fun update() = mainLaunch {
                // Update playback state, position and duration and call super.update() after that.
                playbackState = buildPlaybackState(player.playWhenReady, player.playbackState)
                duration = player.duration.coerceAtLeast(-1)
                position = player.currentPosition.coerceIn(0L..max(0L, duration))
                super.update()
            }.asUnit()

            private fun buildPlaybackState(playWhenReady: Boolean, playbackState: Int): PlaybackState = when (playbackState) {
                Player.STATE_BUFFERING -> PlaybackState.BUFFERING
                Player.STATE_READY -> if (playWhenReady) PlaybackState.PLAYING else PlaybackState.PAUSED
                Player.STATE_ENDED -> PlaybackState.ENDED
                else -> PlaybackState.IDLE
            }

            override fun loadAd(url: String) {
                super.loadAd(url)
                updateJobActive = true

                // Remember the playback position of the content video before switching to the ad.
                if (isShowingContent) {
                    lastContentVideoWindowIndex = player.currentWindowIndex
                    lastContentVideoPosition = player.currentPosition
                }
                isShowingContent = false
                loadMedia(url, autoPlay = true, disableSeeking = true)
            }

            override suspend fun loadContent() {
                updateJobActive = true

                if (isShowingContent) {
                    Log.w("showContent called when already showing content")
                    return
                }
                // Restore any stored playback position when returning from an ad.
                isShowingContent = true
                if (lastContentVideoUrl != contentVideoUrl) {
                    lastContentVideoPosition = null
                    lastContentVideoWindowIndex = null
                }
                lastContentVideoUrl = contentVideoUrl
                loadMedia(contentVideoUrl, lastContentVideoPosition,
                        lastContentVideoWindowIndex, autoPlay = true, disableSeeking = false)
            }

            private fun loadMedia(mediaUrl: String,
                                  initPosition: Long? = null,
                                  initWindowIndex: Int? = null,
                                  autoPlay: Boolean = true,
                                  disableSeeking: Boolean = true) {
                Log.i("videoUri=$mediaUrl; disableSeeking=$disableSeeking; " +
                        "initPosition=$initPosition; initWindowIndex=$initWindowIndex")

                val videoSource = ProgressiveMediaSource.Factory(dataSourceFactory)
                        .createMediaSource(Uri.parse(mediaUrl))
                player.prepare(videoSource, initPosition == null, true)
                initPosition?.also { position ->
                    initWindowIndex?.also { windowIndex ->
                        player.seekTo(windowIndex, position)
                    } ?: player.seekTo(position)
                }
                player.playWhenReady = autoPlay
                playerView.useController = !disableSeeking
            }

            override fun seek(offset: Long) =
                    mainLaunch { player.seekTo(player.currentPosition + offset) }.asUnit()

            override suspend fun sequenceFinished() = playerView.toGone().asUnit()

            override fun release() {
                player.removeListener(exoListener)
                player.release()
                mainScope.cancel()
            }
        }
    }

    private val mainScope: CoroutineScope = object : CoroutineScope {
        override val coroutineContext: CoroutineContext = Dispatchers.Main + Job()
    }

    private lateinit var playerWrapper: SxInstreamVideoPlayerWrapper
    private lateinit var sequencer: SxSequencer

    private val adSlots: List<SxSequencerAdSlot>
        get() = listOf(
                createAdSlot(0.0),
                createAdSlot(0.33),
                createAdSlot(0.66),
                createAdSlot(1.0)
        )

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        setContentView(R.layout.activity_in_stream)

        playerWrapper = ReferenceVideoPlayerWrapper(video,
                getString(R.string.demo_video_short),
                "Some user agent")
        sequencer = SxSequencer(playerWrapper, adSlots, getConfiguration())
        playerWrapper.playWhenReady = true

        title = getString(R.string.title_instream_custom)
    }

    private fun createAdSlot(position: Double) =
            SxSequencerAdSlot(getString(R.string.ad_tag_short),
                    SxSequencerRelativePosition(position),
                    getString(R.string.default_opener_url),
                    false,
                    getString(R.string.default_closer_url),
                    false,
                    getString(R.string.default_bumper_url),
                    false)


    private fun getConfiguration() = SxConfiguration.INSTREAM.apply {
        layoutConfiguration = SxLayoutConfiguration(
                width = ViewGroup.LayoutParams.MATCH_PARENT,
                height = 0,
                orientation = SxLayoutConfiguration.SxOrientation.HORIZONTAL
        )
        clickThroughListener = { callback ->
            mainScope.launch {
                AlertDialog.Builder(this@AbsVideoPlayerWrapperActivity)
                        .setTitle("Custom Click Through Dialog")
                        .setMessage("Do you want to proceed?")
                        .setPositiveButton("Yes") { _, _ -> callback(true) }
                        .setNegativeButton("No") { _, _ -> callback(false) }
                        .setOnCancelListener { callback(false) }
                        .show()
            }
            true
        }
    }

    override fun onResume() {
        super.onResume()
        playerWrapper.playWhenReady = true
    }

    override fun onPause() {
        playerWrapper.playWhenReady = false
        super.onPause()
    }

    override fun onDestroy() {
        sequencer.release()
        super.onDestroy()
    }
}